﻿using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace NetTest
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        // キーボード、コントローラ状態の宣言
        KeyboardState keyboardState;
        GamePadState gamePadState;
        
        // フォントの宣言
        SpriteFont arial;

        // ネットワークセッションの宣言
        NetworkSession networkSession;
        int maxLocalGamers = 4;
        int maxGamers = 16;

        // パケットの宣言
        PacketWriter packetWriter = new PacketWriter();
        PacketReader packetReader = new PacketReader();
        
        // 出力メッセージ
        string message = "sign in...";
        string sendMessage = "--";
        string readMessage = "--";

        int playerDir = 0;
        bool leftCheck = false;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            // コンポーネントの追加
            Components.Add(new GamerServicesComponent(this));
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // TODO: Add your initialization logic here

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            // TODO: use this.Content to load your game content here

            // フォントの呼出
            arial = Content.Load<SpriteFont>("Arial");
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // キーボード、コントローラ状態の取得
            keyboardState = Keyboard.GetState();
            gamePadState = GamePad.GetState(PlayerIndex.One);

            // Allows the game to exit
            if ((gamePadState.Buttons.Back == ButtonState.Pressed)||(keyboardState.IsKeyDown(Keys.Escape)))
            {
                // セッションが残っていたら切断
                if (networkSession != null)
                {
                    networkSession.Dispose();
                    networkSession = null;
                }
                leftCheck = true;
                // 終了
                this.Exit();
            }

            // TODO: Add your update logic here

            // セッションの確認
            if (networkSession == null)
            {
                if ((IsActive)&&(!leftCheck))
                {
                    if (Gamer.SignedInGamers.Count == 0)
                    {
                        // サインインしていなかったらサインイン画面を表示
                        Guide.ShowSignIn(maxLocalGamers, false);
                    }
                    else
                    {
                        try
                        {
                            // セッションを探す
                            using (AvailableNetworkSessionCollection availableSessions =
                                        NetworkSession.Find(NetworkSessionType.SystemLink,
                                                            maxLocalGamers, null))
                            {
                                // セッションが見つからない
                                if (availableSessions.Count == 0)
                                {
                                    try
                                    {
                                        // セッションの作成
                                        networkSession = NetworkSession.Create(NetworkSessionType.SystemLink,
                                                                               maxLocalGamers, maxGamers);

                                        message = "Create Session";
                                    }
                                    catch (Exception e)
                                    {
                                        message = e.Message;
                                    }
                                }
                                else
                                {

                                    // セッションへ接続
                                    networkSession = NetworkSession.Join(availableSessions[0]);
                                    message = "Join Session";
                                }
                                // 参加イベントの追加
                                networkSession.GamerJoined += new EventHandler<GamerJoinedEventArgs>(SessionJoin);
                                // 離脱イベントの追加
                                networkSession.GamerLeft += new EventHandler<GamerLeftEventArgs>(SessionLeft);
                                // 切断イベントの追加
                                networkSession.SessionEnded += new EventHandler<NetworkSessionEndedEventArgs>(SessionEnd);
                            }
                        }
                        catch (Exception e)
                        {
                            message = e.Message;
                        }
                    }
                }
            }
            else
            {
                // 移動（キーボード）
                if (keyboardState.IsKeyDown(Keys.Down)) playerDir = 0;
                if (keyboardState.IsKeyDown(Keys.Left)) playerDir = 1;
                if (keyboardState.IsKeyDown(Keys.Right)) playerDir = 2;
                if (keyboardState.IsKeyDown(Keys.Up)) playerDir = 3;
                // 移動（コントローラ）
                if (gamePadState.DPad.Down == ButtonState.Pressed) playerDir = 0;
                if (gamePadState.DPad.Left == ButtonState.Pressed) playerDir = 1;
                if (gamePadState.DPad.Right == ButtonState.Pressed) playerDir = 2;
                if (gamePadState.DPad.Up == ButtonState.Pressed) playerDir = 3;

                // パケットの送信
                foreach (LocalNetworkGamer gamer in networkSession.LocalGamers)
                {
                    // 移動処理
                    Character player = gamer.Tag as Character;
                    player.Dir = playerDir;
                    player.Update();

                    // 移動方向をパケットに書き込む
                    packetWriter.Write(player.Dir);
                    // 移動先をパケットに書き込む
                    packetWriter.Write(player.Position);

                    // 全接続先へパケットを送信
                    gamer.SendData(packetWriter, SendDataOptions.InOrder);

                    // 送信パケット内容
                    sendMessage = "Direction: " + player.Dir + ", Position: " + player.Position.ToString();
                }

                // セッションの更新
                networkSession.Update();

                // セッションが切断していたら戻る
                if (networkSession == null)
                    return;

                // パケットの読み込み
                foreach (LocalNetworkGamer gamer in networkSession.LocalGamers)
                {
                    while (gamer.IsDataAvailable)
                    {
                        NetworkGamer sender;

                        // データの受信
                        gamer.ReceiveData(packetReader, out sender);

                        // 自分自身からの送信の場合はスキップ
                        if (sender.IsLocal)
                            continue;

                        // 送り主のキャラクタの呼出
                        Character senderPlayer = sender.Tag as Character;

                        // パケットから移動方向を読み込む                                
                        senderPlayer.Dir = packetReader.ReadInt32();
                        // パケットから移動先を読み込む                                
                        senderPlayer.Position = packetReader.ReadVector2();

                        // 受信パケット内容
                        readMessage = "Sender: " + sender.Gamertag + ", Direction: " + senderPlayer.Dir + ", Position: " + senderPlayer.Position.ToString();
                    }
                }
            }

            base.Update(gameTime);
        }

        // セッション参加処理
        public void SessionJoin(object sender, GamerJoinedEventArgs e)
        {
            // セッション参加メッセージ
            message += "\nJoin: " + e.Gamer.Gamertag;
            // キャラクタの生成
            e.Gamer.Tag = new Character(Content);
        }

        // セッション離脱処理
        public void SessionLeft(object sender, GamerLeftEventArgs e)
        {
            // セッション離脱メッセージ
            message += "\nLeft: " + e.Gamer.Gamertag;
        }

        // セッション終了処理
        public void SessionEnd(object sender, NetworkSessionEndedEventArgs e)
        {
            // セッション終了メッセージ
            message = e.EndReason.ToString();
            // 切断処理
            networkSession.Dispose();
            networkSession = null;
        }


        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            graphics.GraphicsDevice.Clear(Color.CornflowerBlue);

            // TODO: Add your drawing code here
            spriteBatch.Begin();

            // メッセージの表示
            spriteBatch.DrawString(arial, message, new Vector2(100, 100), Color.White);

            if (networkSession != null)
            {
                // パケット内容の表示
                spriteBatch.DrawString(arial, "send: " + sendMessage, new Vector2(100, 520), Color.White);
                spriteBatch.DrawString(arial, "read: " + readMessage, new Vector2(100, 540), Color.White);

                // 接続しているキャラクタの描画
                foreach (NetworkGamer gamer in networkSession.AllGamers)
                {
                    // キャラクターの描画
                    Character player = gamer.Tag as Character;
                    player.Draw(spriteBatch, gameTime);

                    // ゲーマータグの表示
                    if (gamer.IsLocal)
                    {
                        // 自分を赤で表示
                        spriteBatch.DrawString(arial, gamer.Gamertag, player.Position + new Vector2(0, -20), Color.Red);
                    }
                    else
                    {
                        // 自分以外を青で表示
                        spriteBatch.DrawString(arial, gamer.Gamertag, player.Position + new Vector2(0, -20), Color.Blue);
                    }
                }

            }
            spriteBatch.End();

            base.Draw(gameTime);
        }
    }
}
